# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/stats.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/stats.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import cards_pb2 as anki_dot_cards__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61nki/stats.proto\x12\nanki.stats\x1a\x12\x61nki/generic.proto\x1a\x10\x61nki/cards.proto\"M\n\nReviewLogs\x12?\n\x07\x65ntries\x18\x01 \x03(\x0b\x32..anki.stats.CardStatsResponse.StatsRevlogEntry\"\xdd\x07\n\x11\x43\x61rdStatsResponse\x12>\n\x06revlog\x18\x01 \x03(\x0b\x32..anki.stats.CardStatsResponse.StatsRevlogEntry\x12\x0f\n\x07\x63\x61rd_id\x18\x02 \x01(\x03\x12\x0f\n\x07note_id\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x65\x63k\x18\x04 \x01(\t\x12\r\n\x05\x61\x64\x64\x65\x64\x18\x05 \x01(\x03\x12\x19\n\x0c\x66irst_review\x18\x06 \x01(\x03H\x00\x88\x01\x01\x12\x1a\n\rlatest_review\x18\x07 \x01(\x03H\x01\x88\x01\x01\x12\x15\n\x08\x64ue_date\x18\x08 \x01(\x03H\x02\x88\x01\x01\x12\x19\n\x0c\x64ue_position\x18\t \x01(\x05H\x03\x88\x01\x01\x12\x10\n\x08interval\x18\n \x01(\r\x12\x0c\n\x04\x65\x61se\x18\x0b \x01(\r\x12\x0f\n\x07reviews\x18\x0c \x01(\r\x12\x0e\n\x06lapses\x18\r \x01(\r\x12\x14\n\x0c\x61verage_secs\x18\x0e \x01(\x02\x12\x12\n\ntotal_secs\x18\x0f \x01(\x02\x12\x11\n\tcard_type\x18\x10 \x01(\t\x12\x10\n\x08notetype\x18\x11 \x01(\t\x12\x36\n\x0cmemory_state\x18\x12 \x01(\x0b\x32\x1b.anki.cards.FsrsMemoryStateH\x04\x88\x01\x01\x12 \n\x13\x66srs_retrievability\x18\x13 \x01(\x02H\x05\x88\x01\x01\x12\x13\n\x0b\x63ustom_data\x18\x14 \x01(\t\x12\x0e\n\x06preset\x18\x15 \x01(\t\x12\x1a\n\roriginal_deck\x18\x16 \x01(\tH\x06\x88\x01\x01\x12\x1e\n\x11\x64\x65sired_retention\x18\x17 \x01(\x02H\x07\x88\x01\x01\x12\x13\n\x0b\x66srs_params\x18\x18 \x03(\x02\x1a\xed\x01\n\x10StatsRevlogEntry\x12\x0c\n\x04time\x18\x01 \x01(\x03\x12\x37\n\x0breview_kind\x18\x02 \x01(\x0e\x32\".anki.stats.RevlogEntry.ReviewKind\x12\x15\n\rbutton_chosen\x18\x03 \x01(\r\x12\x10\n\x08interval\x18\x04 \x01(\r\x12\x0c\n\x04\x65\x61se\x18\x05 \x01(\r\x12\x12\n\ntaken_secs\x18\x06 \x01(\x02\x12\x36\n\x0cmemory_state\x18\x07 \x01(\x0b\x32\x1b.anki.cards.FsrsMemoryStateH\x00\x88\x01\x01\x42\x0f\n\r_memory_stateB\x0f\n\r_first_reviewB\x10\n\x0e_latest_reviewB\x0b\n\t_due_dateB\x0f\n\r_due_positionB\x0f\n\r_memory_stateB\x16\n\x14_fsrs_retrievabilityB\x10\n\x0e_original_deckB\x14\n\x12_desired_retention\"-\n\rGraphsRequest\x12\x0e\n\x06search\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ys\x18\x02 \x01(\r\"\xc0\x1d\n\x0eGraphsResponse\x12\x33\n\x07\x62uttons\x18\x01 \x01(\x0b\x32\".anki.stats.GraphsResponse.Buttons\x12:\n\x0b\x63\x61rd_counts\x18\x02 \x01(\x0b\x32%.anki.stats.GraphsResponse.CardCounts\x12/\n\x05hours\x18\x03 \x01(\x0b\x32 .anki.stats.GraphsResponse.Hours\x12/\n\x05today\x18\x04 \x01(\x0b\x32 .anki.stats.GraphsResponse.Today\x12/\n\x05\x65\x61ses\x18\x05 \x01(\x0b\x32 .anki.stats.GraphsResponse.Eases\x12\x34\n\ndifficulty\x18\x0b \x01(\x0b\x32 .anki.stats.GraphsResponse.Eases\x12\x37\n\tintervals\x18\x06 \x01(\x0b\x32$.anki.stats.GraphsResponse.Intervals\x12\x38\n\nfuture_due\x18\x07 \x01(\x0b\x32$.anki.stats.GraphsResponse.FutureDue\x12/\n\x05\x61\x64\x64\x65\x64\x18\x08 \x01(\x0b\x32 .anki.stats.GraphsResponse.Added\x12@\n\x07reviews\x18\t \x01(\x0b\x32/.anki.stats.GraphsResponse.ReviewCountsAndTimes\x12\x15\n\rrollover_hour\x18\n \x01(\r\x12\x41\n\x0eretrievability\x18\x0c \x01(\x0b\x32).anki.stats.GraphsResponse.Retrievability\x12\x0c\n\x04\x66srs\x18\r \x01(\x08\x12\x37\n\tstability\x18\x0e \x01(\x0b\x32$.anki.stats.GraphsResponse.Intervals\x12\x45\n\x0etrue_retention\x18\x0f \x01(\x0b\x32-.anki.stats.GraphsResponse.TrueRetentionStats\x1aq\n\x05\x41\x64\x64\x65\x64\x12:\n\x05\x61\x64\x64\x65\x64\x18\x01 \x03(\x0b\x32+.anki.stats.GraphsResponse.Added.AddedEntry\x1a,\n\nAddedEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\x85\x01\n\tIntervals\x12\x46\n\tintervals\x18\x01 \x03(\x0b\x32\x33.anki.stats.GraphsResponse.Intervals.IntervalsEntry\x1a\x30\n\x0eIntervalsEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\x82\x01\n\x05\x45\x61ses\x12:\n\x05\x65\x61ses\x18\x01 \x03(\x0b\x32+.anki.stats.GraphsResponse.Eases.EasesEntry\x12\x0f\n\x07\x61verage\x18\x02 \x01(\x02\x1a,\n\nEasesEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\xd9\x01\n\x0eRetrievability\x12U\n\x0eretrievability\x18\x01 \x03(\x0b\x32=.anki.stats.GraphsResponse.Retrievability.RetrievabilityEntry\x12\x0f\n\x07\x61verage\x18\x02 \x01(\x02\x12\x13\n\x0bsum_by_card\x18\x03 \x01(\x02\x12\x13\n\x0bsum_by_note\x18\x04 \x01(\x02\x1a\x35\n\x13RetrievabilityEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\xb0\x01\n\tFutureDue\x12G\n\nfuture_due\x18\x01 \x03(\x0b\x32\x33.anki.stats.GraphsResponse.FutureDue.FutureDueEntry\x12\x14\n\x0chave_backlog\x18\x02 \x01(\x08\x12\x12\n\ndaily_load\x18\x03 \x01(\r\x1a\x30\n\x0e\x46utureDueEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\x1a\xd7\x01\n\x05Today\x12\x14\n\x0c\x61nswer_count\x18\x01 \x01(\r\x12\x15\n\ranswer_millis\x18\x02 \x01(\r\x12\x15\n\rcorrect_count\x18\x03 \x01(\r\x12\x16\n\x0emature_correct\x18\x04 \x01(\r\x12\x14\n\x0cmature_count\x18\x05 \x01(\r\x12\x13\n\x0blearn_count\x18\x06 \x01(\r\x12\x14\n\x0creview_count\x18\x07 \x01(\r\x12\x15\n\rrelearn_count\x18\x08 \x01(\r\x12\x1a\n\x12\x65\x61rly_review_count\x18\t \x01(\r\x1a\x98\x02\n\x05Hours\x12\x38\n\tone_month\x18\x01 \x03(\x0b\x32%.anki.stats.GraphsResponse.Hours.Hour\x12;\n\x0cthree_months\x18\x02 \x03(\x0b\x32%.anki.stats.GraphsResponse.Hours.Hour\x12\x37\n\x08one_year\x18\x03 \x03(\x0b\x32%.anki.stats.GraphsResponse.Hours.Hour\x12\x37\n\x08\x61ll_time\x18\x04 \x03(\x0b\x32%.anki.stats.GraphsResponse.Hours.Hour\x1a&\n\x04Hour\x12\r\n\x05total\x18\x01 \x01(\r\x12\x0f\n\x07\x63orrect\x18\x02 \x01(\r\x1a\xd3\x03\n\x14ReviewCountsAndTimes\x12I\n\x05\x63ount\x18\x01 \x03(\x0b\x32:.anki.stats.GraphsResponse.ReviewCountsAndTimes.CountEntry\x12G\n\x04time\x18\x02 \x03(\x0b\x32\x39.anki.stats.GraphsResponse.ReviewCountsAndTimes.TimeEntry\x1aZ\n\x07Reviews\x12\r\n\x05learn\x18\x01 \x01(\r\x12\x0f\n\x07relearn\x18\x02 \x01(\r\x12\r\n\x05young\x18\x03 \x01(\r\x12\x0e\n\x06mature\x18\x04 \x01(\r\x12\x10\n\x08\x66iltered\x18\x05 \x01(\r\x1a\x65\n\nCountEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x46\n\x05value\x18\x02 \x01(\x0b\x32\x37.anki.stats.GraphsResponse.ReviewCountsAndTimes.Reviews:\x02\x38\x01\x1a\x64\n\tTimeEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x46\n\x05value\x18\x02 \x01(\x0b\x32\x37.anki.stats.GraphsResponse.ReviewCountsAndTimes.Reviews:\x02\x38\x01\x1a\xdb\x02\n\x07\x42uttons\x12\x42\n\tone_month\x18\x01 \x01(\x0b\x32/.anki.stats.GraphsResponse.Buttons.ButtonCounts\x12\x45\n\x0cthree_months\x18\x02 \x01(\x0b\x32/.anki.stats.GraphsResponse.Buttons.ButtonCounts\x12\x41\n\x08one_year\x18\x03 \x01(\x0b\x32/.anki.stats.GraphsResponse.Buttons.ButtonCounts\x12\x41\n\x08\x61ll_time\x18\x04 \x01(\x0b\x32/.anki.stats.GraphsResponse.Buttons.ButtonCounts\x1a?\n\x0c\x42uttonCounts\x12\x10\n\x08learning\x18\x01 \x03(\r\x12\r\n\x05young\x18\x02 \x03(\r\x12\x0e\n\x06mature\x18\x03 \x03(\r\x1a\x9e\x02\n\nCardCounts\x12H\n\x12including_inactive\x18\x01 \x01(\x0b\x32,.anki.stats.GraphsResponse.CardCounts.Counts\x12H\n\x12\x65xcluding_inactive\x18\x02 \x01(\x0b\x32,.anki.stats.GraphsResponse.CardCounts.Counts\x1a|\n\x06\x43ounts\x12\x10\n\x08newCards\x18\x01 \x01(\r\x12\r\n\x05learn\x18\x02 \x01(\r\x12\x0f\n\x07relearn\x18\x03 \x01(\r\x12\r\n\x05young\x18\x04 \x01(\r\x12\x0e\n\x06mature\x18\x05 \x01(\r\x12\x11\n\tsuspended\x18\x06 \x01(\r\x12\x0e\n\x06\x62uried\x18\x07 \x01(\r\x1a\xcc\x04\n\x12TrueRetentionStats\x12J\n\x05today\x18\x01 \x01(\x0b\x32;.anki.stats.GraphsResponse.TrueRetentionStats.TrueRetention\x12N\n\tyesterday\x18\x02 \x01(\x0b\x32;.anki.stats.GraphsResponse.TrueRetentionStats.TrueRetention\x12I\n\x04week\x18\x03 \x01(\x0b\x32;.anki.stats.GraphsResponse.TrueRetentionStats.TrueRetention\x12J\n\x05month\x18\x04 \x01(\x0b\x32;.anki.stats.GraphsResponse.TrueRetentionStats.TrueRetention\x12I\n\x04year\x18\x05 \x01(\x0b\x32;.anki.stats.GraphsResponse.TrueRetentionStats.TrueRetention\x12M\n\x08\x61ll_time\x18\x06 \x01(\x0b\x32;.anki.stats.GraphsResponse.TrueRetentionStats.TrueRetention\x1ai\n\rTrueRetention\x12\x14\n\x0cyoung_passed\x18\x01 \x01(\r\x12\x14\n\x0cyoung_failed\x18\x02 \x01(\r\x12\x15\n\rmature_passed\x18\x03 \x01(\r\x12\x15\n\rmature_failed\x18\x04 \x01(\r\"\x82\x02\n\x10GraphPreferences\x12H\n\x1a\x63\x61lendar_first_day_of_week\x18\x01 \x01(\x0e\x32$.anki.stats.GraphPreferences.Weekday\x12%\n\x1d\x63\x61rd_counts_separate_inactive\x18\x02 \x01(\x08\x12\x1f\n\x17\x62rowser_links_supported\x18\x03 \x01(\x08\x12\x1f\n\x17\x66uture_due_show_backlog\x18\x04 \x01(\x08\";\n\x07Weekday\x12\n\n\x06SUNDAY\x10\x00\x12\n\n\x06MONDAY\x10\x01\x12\n\n\x06\x46RIDAY\x10\x05\x12\x0c\n\x08SATURDAY\x10\x06\"\xba\x02\n\x0bRevlogEntry\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0b\n\x03\x63id\x18\x02 \x01(\x03\x12\x0b\n\x03usn\x18\x03 \x01(\x05\x12\x15\n\rbutton_chosen\x18\x04 \x01(\r\x12\x10\n\x08interval\x18\x05 \x01(\x05\x12\x15\n\rlast_interval\x18\x06 \x01(\x05\x12\x13\n\x0b\x65\x61se_factor\x18\x07 \x01(\r\x12\x14\n\x0ctaken_millis\x18\x08 \x01(\r\x12\x37\n\x0breview_kind\x18\t \x01(\x0e\x32\".anki.stats.RevlogEntry.ReviewKind\"a\n\nReviewKind\x12\x0c\n\x08LEARNING\x10\x00\x12\n\n\x06REVIEW\x10\x01\x12\x0e\n\nRELEARNING\x10\x02\x12\x0c\n\x08\x46ILTERED\x10\x03\x12\n\n\x06MANUAL\x10\x04\x12\x0f\n\x0bRESCHEDULED\x10\x05\"9\n\tCardEntry\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0f\n\x07note_id\x18\x02 \x01(\x03\x12\x0f\n\x07\x64\x65\x63k_id\x18\x03 \x01(\x03\"=\n\tDeckEntry\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x11\n\tparent_id\x18\x02 \x01(\x03\x12\x11\n\tpreset_id\x18\x03 \x01(\x03\"\x94\x01\n\x07\x44\x61taset\x12(\n\x07revlogs\x18\x01 \x03(\x0b\x32\x17.anki.stats.RevlogEntry\x12$\n\x05\x63\x61rds\x18\x02 \x03(\x0b\x32\x15.anki.stats.CardEntry\x12$\n\x05\x64\x65\x63ks\x18\x03 \x03(\x0b\x32\x15.anki.stats.DeckEntry\x12\x13\n\x0bnext_day_at\x18\x04 \x01(\x03\x32\xe0\x02\n\x0cStatsService\x12>\n\tCardStats\x12\x12.anki.cards.CardId\x1a\x1d.anki.stats.CardStatsResponse\x12;\n\rGetReviewLogs\x12\x12.anki.cards.CardId\x1a\x16.anki.stats.ReviewLogs\x12?\n\x06Graphs\x12\x19.anki.stats.GraphsRequest\x1a\x1a.anki.stats.GraphsResponse\x12H\n\x13GetGraphPreferences\x12\x13.anki.generic.Empty\x1a\x1c.anki.stats.GraphPreferences\x12H\n\x13SetGraphPreferences\x12\x1c.anki.stats.GraphPreferences\x1a\x13.anki.generic.Empty2\x15\n\x13\x42\x61\x63kendStatsServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.stats_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_GRAPHSRESPONSE_ADDED_ADDEDENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_ADDED_ADDEDENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHSRESPONSE_INTERVALS_INTERVALSENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_INTERVALS_INTERVALSENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHSRESPONSE_EASES_EASESENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_EASES_EASESENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHSRESPONSE_RETRIEVABILITY_RETRIEVABILITYENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_RETRIEVABILITY_RETRIEVABILITYENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHSRESPONSE_FUTUREDUE_FUTUREDUEENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_FUTUREDUE_FUTUREDUEENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_COUNTENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_COUNTENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_TIMEENTRY']._loaded_options = None
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_TIMEENTRY']._serialized_options = b'8\001'
  _globals['_REVIEWLOGS']._serialized_start=70
  _globals['_REVIEWLOGS']._serialized_end=147
  _globals['_CARDSTATSRESPONSE']._serialized_start=150
  _globals['_CARDSTATSRESPONSE']._serialized_end=1139
  _globals['_CARDSTATSRESPONSE_STATSREVLOGENTRY']._serialized_start=756
  _globals['_CARDSTATSRESPONSE_STATSREVLOGENTRY']._serialized_end=993
  _globals['_GRAPHSREQUEST']._serialized_start=1141
  _globals['_GRAPHSREQUEST']._serialized_end=1186
  _globals['_GRAPHSRESPONSE']._serialized_start=1189
  _globals['_GRAPHSRESPONSE']._serialized_end=4965
  _globals['_GRAPHSRESPONSE_ADDED']._serialized_start=1983
  _globals['_GRAPHSRESPONSE_ADDED']._serialized_end=2096
  _globals['_GRAPHSRESPONSE_ADDED_ADDEDENTRY']._serialized_start=2052
  _globals['_GRAPHSRESPONSE_ADDED_ADDEDENTRY']._serialized_end=2096
  _globals['_GRAPHSRESPONSE_INTERVALS']._serialized_start=2099
  _globals['_GRAPHSRESPONSE_INTERVALS']._serialized_end=2232
  _globals['_GRAPHSRESPONSE_INTERVALS_INTERVALSENTRY']._serialized_start=2184
  _globals['_GRAPHSRESPONSE_INTERVALS_INTERVALSENTRY']._serialized_end=2232
  _globals['_GRAPHSRESPONSE_EASES']._serialized_start=2235
  _globals['_GRAPHSRESPONSE_EASES']._serialized_end=2365
  _globals['_GRAPHSRESPONSE_EASES_EASESENTRY']._serialized_start=2321
  _globals['_GRAPHSRESPONSE_EASES_EASESENTRY']._serialized_end=2365
  _globals['_GRAPHSRESPONSE_RETRIEVABILITY']._serialized_start=2368
  _globals['_GRAPHSRESPONSE_RETRIEVABILITY']._serialized_end=2585
  _globals['_GRAPHSRESPONSE_RETRIEVABILITY_RETRIEVABILITYENTRY']._serialized_start=2532
  _globals['_GRAPHSRESPONSE_RETRIEVABILITY_RETRIEVABILITYENTRY']._serialized_end=2585
  _globals['_GRAPHSRESPONSE_FUTUREDUE']._serialized_start=2588
  _globals['_GRAPHSRESPONSE_FUTUREDUE']._serialized_end=2764
  _globals['_GRAPHSRESPONSE_FUTUREDUE_FUTUREDUEENTRY']._serialized_start=2716
  _globals['_GRAPHSRESPONSE_FUTUREDUE_FUTUREDUEENTRY']._serialized_end=2764
  _globals['_GRAPHSRESPONSE_TODAY']._serialized_start=2767
  _globals['_GRAPHSRESPONSE_TODAY']._serialized_end=2982
  _globals['_GRAPHSRESPONSE_HOURS']._serialized_start=2985
  _globals['_GRAPHSRESPONSE_HOURS']._serialized_end=3265
  _globals['_GRAPHSRESPONSE_HOURS_HOUR']._serialized_start=3227
  _globals['_GRAPHSRESPONSE_HOURS_HOUR']._serialized_end=3265
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES']._serialized_start=3268
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES']._serialized_end=3735
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_REVIEWS']._serialized_start=3440
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_REVIEWS']._serialized_end=3530
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_COUNTENTRY']._serialized_start=3532
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_COUNTENTRY']._serialized_end=3633
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_TIMEENTRY']._serialized_start=3635
  _globals['_GRAPHSRESPONSE_REVIEWCOUNTSANDTIMES_TIMEENTRY']._serialized_end=3735
  _globals['_GRAPHSRESPONSE_BUTTONS']._serialized_start=3738
  _globals['_GRAPHSRESPONSE_BUTTONS']._serialized_end=4085
  _globals['_GRAPHSRESPONSE_BUTTONS_BUTTONCOUNTS']._serialized_start=4022
  _globals['_GRAPHSRESPONSE_BUTTONS_BUTTONCOUNTS']._serialized_end=4085
  _globals['_GRAPHSRESPONSE_CARDCOUNTS']._serialized_start=4088
  _globals['_GRAPHSRESPONSE_CARDCOUNTS']._serialized_end=4374
  _globals['_GRAPHSRESPONSE_CARDCOUNTS_COUNTS']._serialized_start=4250
  _globals['_GRAPHSRESPONSE_CARDCOUNTS_COUNTS']._serialized_end=4374
  _globals['_GRAPHSRESPONSE_TRUERETENTIONSTATS']._serialized_start=4377
  _globals['_GRAPHSRESPONSE_TRUERETENTIONSTATS']._serialized_end=4965
  _globals['_GRAPHSRESPONSE_TRUERETENTIONSTATS_TRUERETENTION']._serialized_start=4860
  _globals['_GRAPHSRESPONSE_TRUERETENTIONSTATS_TRUERETENTION']._serialized_end=4965
  _globals['_GRAPHPREFERENCES']._serialized_start=4968
  _globals['_GRAPHPREFERENCES']._serialized_end=5226
  _globals['_GRAPHPREFERENCES_WEEKDAY']._serialized_start=5167
  _globals['_GRAPHPREFERENCES_WEEKDAY']._serialized_end=5226
  _globals['_REVLOGENTRY']._serialized_start=5229
  _globals['_REVLOGENTRY']._serialized_end=5543
  _globals['_REVLOGENTRY_REVIEWKIND']._serialized_start=5446
  _globals['_REVLOGENTRY_REVIEWKIND']._serialized_end=5543
  _globals['_CARDENTRY']._serialized_start=5545
  _globals['_CARDENTRY']._serialized_end=5602
  _globals['_DECKENTRY']._serialized_start=5604
  _globals['_DECKENTRY']._serialized_end=5665
  _globals['_DATASET']._serialized_start=5668
  _globals['_DATASET']._serialized_end=5816
  _globals['_STATSSERVICE']._serialized_start=5819
  _globals['_STATSSERVICE']._serialized_end=6171
  _globals['_BACKENDSTATSSERVICE']._serialized_start=6173
  _globals['_BACKENDSTATSSERVICE']._serialized_end=6194
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

# ruff: noqa: F401
import copy
import time
from collections.abc import Generator, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

import aqt
import aqt.browser
from anki.cards import CardId
from anki.collection import BrowserColumns as Columns
from anki.collection import BrowserRow
from anki.notes import NoteId
from aqt import colors
from aqt.qt import QColor
from aqt.utils import tr

Column = Columns.Column
ItemId = Union[CardId, NoteId]
ItemList = Union[Sequence[CardId], Sequence[NoteId]]


@dataclass
class SearchContext:
    search: str
    browser: aqt.browser.Browser
    order: bool | str | Column = True
    reverse: bool = False
    addon_metadata: dict | None = None
    # if set, provided ids will be used instead of the regular search
    ids: Sequence[ItemId] | None = None


class Cell:
    def __init__(
        self, text: str, is_rtl: bool, elide_mode: BrowserRow.Cell.TextElideMode.V
    ) -> None:
        self.text: str = text
        self.is_rtl: bool = is_rtl
        self.elide_mode: aqt.Qt.TextElideMode = backend_elide_mode_to_aqt_elide_mode(
            elide_mode
        )


class CellRow:
    is_disabled: bool = False

    def __init__(
        self,
        cells: Generator[tuple[str, bool, BrowserRow.Cell.TextElideMode.V], None, None],
        color: BrowserRow.Color.V,
        font_name: str,
        font_size: int,
    ) -> None:
        self.refreshed_at: float = time.time()
        self.cells: tuple[Cell, ...] = tuple(Cell(*cell) for cell in cells)
        self.color: dict[str, str] | None = backend_color_to_aqt_color(color)
        self.font_name: str = font_name or "arial"
        self.font_size: int = font_size if font_size > 0 else 12

    def is_stale(self, threshold: float) -> bool:
        return self.refreshed_at < threshold

    @staticmethod
    def generic(length: int, cell_text: str) -> CellRow:
        return CellRow(
            ((cell_text, False, BrowserRow.Cell.ElideRight) for cell in range(length)),
            BrowserRow.COLOR_DEFAULT,
            "arial",
            12,
        )

    @staticmethod
    def placeholder(length: int) -> CellRow:
        return CellRow.generic(length, "...")

    @staticmethod
    def disabled(length: int, cell_text: str) -> CellRow:
        row = CellRow.generic(length, cell_text)
        row.is_disabled = True
        return row


def backend_elide_mode_to_aqt_elide_mode(
    elide_mode: BrowserRow.Cell.TextElideMode.V,
) -> aqt.Qt.TextElideMode:
    if elide_mode == BrowserRow.Cell.ElideLeft:
        return aqt.Qt.TextElideMode.ElideLeft
    if elide_mode == BrowserRow.Cell.ElideMiddle:
        return aqt.Qt.TextElideMode.ElideMiddle
    if elide_mode == BrowserRow.Cell.ElideNone:
        return aqt.Qt.TextElideMode.ElideNone
    return aqt.Qt.TextElideMode.ElideRight


def backend_color_to_aqt_color(color: BrowserRow.Color.V) -> dict[str, str] | None:
    temp_color = None

    if color == BrowserRow.COLOR_MARKED:
        temp_color = colors.STATE_MARKED
    if color == BrowserRow.COLOR_SUSPENDED:
        temp_color = colors.STATE_SUSPENDED
    if color == BrowserRow.COLOR_BURIED:
        temp_color = colors.STATE_BURIED
    if color == BrowserRow.COLOR_FLAG_RED:
        temp_color = colors.FLAG_1
    if color == BrowserRow.COLOR_FLAG_ORANGE:
        temp_color = colors.FLAG_2
    if color == BrowserRow.COLOR_FLAG_GREEN:
        temp_color = colors.FLAG_3
    if color == BrowserRow.COLOR_FLAG_BLUE:
        temp_color = colors.FLAG_4
    if color == BrowserRow.COLOR_FLAG_PINK:
        temp_color = colors.FLAG_5
    if color == BrowserRow.COLOR_FLAG_TURQUOISE:
        temp_color = colors.FLAG_6
    if color == BrowserRow.COLOR_FLAG_PURPLE:
        temp_color = colors.FLAG_7

    return adjusted_bg_color(temp_color)


def adjusted_bg_color(color: dict[str, str] | None) -> dict[str, str] | None:
    if color:
        adjusted_color = copy.copy(color)
        light = QColor(color["light"]).lighter(150)
        adjusted_color["light"] = light.name()
        dark = QColor(color["dark"]).darker(150)
        adjusted_color["dark"] = dark.name()
        return adjusted_color
    else:
        return None


from .model import DataModel
from .state import CardState, ItemState, NoteState
from .table import StatusDelegate, Table



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Sequence

from anki.collection import OpChanges, OpChangesWithCount
from anki.notes import NoteId
from aqt.operations import CollectionOp
from aqt.qt import QWidget
from aqt.utils import showInfo, tooltip, tr


def add_tags_to_notes(
    *,
    parent: QWidget,
    note_ids: Sequence[NoteId],
    space_separated_tags: str,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent, lambda col: col.tags.bulk_add(note_ids, space_separated_tags)
    ).success(
        lambda out: tooltip(tr.browsing_notes_updated(count=out.count), parent=parent)
    )


def remove_tags_from_notes(
    *,
    parent: QWidget,
    note_ids: Sequence[NoteId],
    space_separated_tags: str,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent, lambda col: col.tags.bulk_remove(note_ids, space_separated_tags)
    ).success(
        lambda out: tooltip(tr.browsing_notes_updated(count=out.count), parent=parent)
    )


def clear_unused_tags(*, parent: QWidget) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.tags.clear_unused_tags()).success(
        lambda out: tooltip(
            tr.browsing_removed_unused_tags_count(count=out.count), parent=parent
        )
    )


def rename_tag(
    *,
    parent: QWidget,
    current_name: str,
    new_name: str,
) -> CollectionOp[OpChangesWithCount]:
    def success(out: OpChangesWithCount) -> None:
        if out.count:
            tooltip(tr.browsing_notes_updated(count=out.count), parent=parent)
        else:
            showInfo(tr.browsing_tag_rename_warning_empty(), parent=parent)

    return CollectionOp(
        parent,
        lambda col: col.tags.rename(old=current_name, new=new_name),
    ).success(success)


def remove_tags_from_all_notes(
    *, parent: QWidget, space_separated_tags: str
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent, lambda col: col.tags.remove(space_separated_tags=space_separated_tags)
    ).success(
        lambda out: tooltip(tr.browsing_notes_updated(count=out.count), parent=parent)
    )


def reparent_tags(
    *, parent: QWidget, tags: Sequence[str], new_parent: str
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent, lambda col: col.tags.reparent(tags=tags, new_parent=new_parent)
    ).success(
        lambda out: tooltip(tr.browsing_notes_updated(count=out.count), parent=parent)
    )


def set_tag_collapsed(
    *, parent: QWidget, tag: str, collapsed: bool
) -> CollectionOp[OpChanges]:
    return CollectionOp(
        parent, lambda col: col.tags.set_collapsed(tag=tag, collapsed=collapsed)
    )


def find_and_replace_tag(
    *,
    parent: QWidget,
    note_ids: Sequence[int],
    search: str,
    replacement: str,
    regex: bool,
    match_case: bool,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent,
        lambda col: col.tags.find_and_replace(
            note_ids=note_ids,
            search=search,
            replacement=replacement,
            regex=regex,
            match_case=match_case,
        ),
    ).success(
        lambda out: tooltip(
            tr.findreplace_notes_updated(changed=out.count, total=len(note_ids)),
            parent=parent,
        ),
    )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, NewType

from anki._legacy import DeprecatedNamesMixinForModule

# whether new cards should be mixed with reviews, or shown first or last
NEW_CARDS_DISTRIBUTE = 0
NEW_CARDS_LAST = 1
NEW_CARDS_FIRST = 2

# new card insertion order
NEW_CARDS_RANDOM = 0
NEW_CARDS_DUE = 1

# Queue types
CardQueue = NewType("CardQueue", int)
QUEUE_TYPE_MANUALLY_BURIED = CardQueue(-3)
QUEUE_TYPE_SIBLING_BURIED = CardQueue(-2)
QUEUE_TYPE_SUSPENDED = CardQueue(-1)
QUEUE_TYPE_NEW = CardQueue(0)
QUEUE_TYPE_LRN = CardQueue(1)
QUEUE_TYPE_REV = CardQueue(2)
QUEUE_TYPE_DAY_LEARN_RELEARN = CardQueue(3)
QUEUE_TYPE_PREVIEW = CardQueue(4)

# Card types
CardType = NewType("CardType", int)
CARD_TYPE_NEW = CardType(0)
CARD_TYPE_LRN = CardType(1)
CARD_TYPE_REV = CardType(2)
CARD_TYPE_RELEARNING = CardType(3)

# removal types
REM_CARD = 0
REM_NOTE = 1
REM_DECK = 2

# count display
COUNT_ANSWERED = 0
COUNT_REMAINING = 1

# media log
MEDIA_ADD = 0
MEDIA_REM = 1

# Kind of decks
DECK_STD = 0
DECK_DYN = 1

# dynamic deck order
DYN_OLDEST = 0
DYN_RANDOM = 1
DYN_SMALLINT = 2
DYN_BIGINT = 3
DYN_LAPSES = 4
DYN_ADDED = 5
DYN_DUE = 6
DYN_REVADDED = 7
DYN_DUEPRIORITY = 8

DYN_MAX_SIZE = 99999

# model types
MODEL_STD = 0
MODEL_CLOZE = 1

STARTING_FACTOR = 2500
STARTING_FACTOR_FRACTION = STARTING_FACTOR / 1000

HELP_SITE = "https://docs.ankiweb.net/"

# Leech actions
LEECH_SUSPEND = 0
LEECH_TAGONLY = 1

# Buttons
BUTTON_ONE = 1
BUTTON_TWO = 2
BUTTON_THREE = 3
BUTTON_FOUR = 4

# Revlog types
REVLOG_LRN = 0
REVLOG_REV = 1
REVLOG_RELRN = 2
REVLOG_CRAM = 3
REVLOG_RESCHED = 4

# Labels
##########################################################################

import anki.collection


def _tr(col: anki.collection.Collection | None) -> Any:
    if col:
        return col.tr
    else:
        print("routine in consts.py should be passed col")
        import traceback

        traceback.print_stack(file=sys.stdout)
        from anki.lang import tr_legacyglobal

        return tr_legacyglobal


def new_card_order_labels(col: anki.collection.Collection | None) -> dict[int, Any]:
    tr = _tr(col)
    return {
        0: tr.scheduling_show_new_cards_in_random_order(),
        1: tr.scheduling_show_new_cards_in_order_added(),
    }


_deprecated_names = DeprecatedNamesMixinForModule(globals())


if not TYPE_CHECKING:

    def __getattr__(name: str) -> Any:
        return _deprecated_names.__getattr__(name)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

import html
import unicodedata
from typing import Union

from anki.collection import Collection
from anki.config import Config
from anki.consts import NEW_CARDS_RANDOM, STARTING_FACTOR
from anki.importing.base import Importer
from anki.models import NotetypeId
from anki.notes import NoteId
from anki.utils import (
    field_checksum,
    guid64,
    int_time,
    join_fields,
    split_fields,
    timestamp_id,
)

TagMappedUpdate = tuple[int, int, str, str, NoteId, str, str]
TagModifiedUpdate = tuple[int, int, str, str, NoteId, str]
NoTagUpdate = tuple[int, int, str, NoteId, str]
Updates = Union[TagMappedUpdate, TagModifiedUpdate, NoTagUpdate]

# Stores a list of fields, tags and deck
######################################################################


class ForeignNote:
    "An temporary object storing fields and attributes."

    def __init__(self) -> None:
        self.fields: list[str] = []
        self.tags: list[str] = []
        self.deck = None
        self.cards: dict[int, ForeignCard] = {}  # map of ord -> card
        self.fieldsStr = ""


class ForeignCard:
    def __init__(self) -> None:
        self.due = 0
        self.ivl = 1
        self.factor = STARTING_FACTOR
        self.reps = 0
        self.lapses = 0


# Base class for CSV and similar text-based imports
######################################################################

# The mapping is list of input fields, like:
# ['Expression', 'Reading', '_tags', None]
# - None means that the input should be discarded
# - _tags maps to note tags
# If the first field of the model is not in the map, the map is invalid.

# The import mode is one of:
# UPDATE_MODE: update if first field matches existing note
# IGNORE_MODE: ignore if first field matches existing note
# ADD_MODE: import even if first field matches existing note
UPDATE_MODE = 0
IGNORE_MODE = 1
ADD_MODE = 2


class NoteImporter(Importer):
    needMapper = True
    needDelimiter = False
    allowHTML = False
    importMode = UPDATE_MODE
    mapping: list[str] | None
    tagModified: str | None

    def __init__(self, col: Collection, file: str) -> None:
        Importer.__init__(self, col, file)
        self.model = col.models.current()
        self.mapping = None
        self.tagModified = None
        self._tagsMapped = False

    def run(self) -> None:
        "Import."
        assert self.mapping
        c = self.foreignNotes()
        self.importNotes(c)

    def fields(self) -> int:
        "The number of fields."
        return 0

    def initMapping(self) -> None:
        flds = [f["name"] for f in self.model["flds"]]
        # truncate to provided count
        flds = flds[0 : self.fields()]
        # if there's room left, add tags
        if self.fields() > len(flds):
            flds.append("_tags")
        # and if there's still room left, pad
        flds = flds + [None] * (self.fields() - len(flds))
        self.mapping = flds

    def mappingOk(self) -> bool:
        return self.model["flds"][0]["name"] in self.mapping

    def foreignNotes(self) -> list:
        "Return a list of foreign notes for importing."
        return []

    def importNotes(self, notes: list[ForeignNote]) -> None:
        "Convert each card into a note, apply attributes and add to col."
        if not self.mappingOk():
            raise Exception("mapping not ok")
        # note whether tags are mapped
        self._tagsMapped = False
        for f in self.mapping:
            if f == "_tags":
                self._tagsMapped = True
        # gather checks for duplicate comparison
        csums: dict[str, list[NoteId]] = {}
        for csum, id in self.col.db.execute(
            "select csum, id from notes where mid = ?", self.model["id"]
        ):
            if csum in csums:
                csums[csum].append(id)
            else:
                csums[csum] = [id]
        firsts: dict[str, bool] = {}
        fld0idx = self.mapping.index(self.model["flds"][0]["name"])
        self._fmap = self.col.models.field_map(self.model)
        self._nextID = NoteId(timestamp_id(self.col.db, "notes"))
        # loop through the notes
        updates: list[Updates] = []
        updateLog = []
        new = []
        self._ids: list[NoteId] = []
        self._cards: list[tuple] = []
        dupeCount = 0
        dupes: list[str] = []
        for n in notes:
            for c, field in enumerate(n.fields):
                if not self.allowHTML:
                    n.fields[c] = html.escape(field, quote=False)
                n.fields[c] = field.strip()
                if not self.allowHTML:
                    n.fields[c] = field.replace("\n", "<br>")
            fld0 = unicodedata.normalize("NFC", n.fields[fld0idx])
            # first field must exist
            if not fld0:
                self.log.append(
                    self.col.tr.importing_empty_first_field(val=" ".join(n.fields))
                )
                continue
            csum = field_checksum(fld0)
            # earlier in import?
            if fld0 in firsts and self.importMode != ADD_MODE:
                # duplicates in source file; log and ignore
                self.log.append(self.col.tr.importing_appeared_twice_in_file(val=fld0))
                continue
            firsts[fld0] = True
            # already exists?
            found = False
            if csum in csums:  # type: ignore[comparison-overlap]
                # csum is not a guarantee; have to check
                for id in csums[csum]:  # type: ignore[index]
                    flds = self.col.db.scalar("select flds from notes where id = ?", id)
                    sflds = split_fields(flds)
                    if fld0 == sflds[0]:
                        # duplicate
                        found = True
                        if self.importMode == UPDATE_MODE:
                            data = self.updateData(n, id, sflds)
                            if data:
                                updates.append(data)
                                updateLog.append(
                                    self.col.tr.importing_first_field_matched(val=fld0)
                                )
                                dupeCount += 1
                                found = True
                        elif self.importMode == IGNORE_MODE:
                            dupeCount += 1
                        elif self.importMode == ADD_MODE:
                            # allow duplicates in this case
                            if fld0 not in dupes:
                                # only show message once, no matter how many
                                # duplicates are in the collection already
                                updateLog.append(
                                    self.col.tr.importing_added_duplicate_with_first_field(
                                        val=fld0,
                                    )
                                )
                                dupes.append(fld0)
                            found = False
            # newly add
            if not found:
                new_data = self.newData(n)
                if new_data:
                    new.append(new_data)
                    # note that we've seen this note once already
                    firsts[fld0] = True
        self.addNew(new)
        self.addUpdates(updates)
        # generate cards + update field cache
        self.col.after_note_updates(self._ids, mark_modified=False)
        # apply scheduling updates
        self.updateCards()
        # we randomize or order here, to ensure that siblings
        # have the same due#
        did = self.col.decks.selected()
        conf = self.col.decks.config_dict_for_deck_id(did)
        # in order due?
        if not conf["dyn"] and conf["new"]["order"] == NEW_CARDS_RANDOM:
            self.col.sched.randomize_cards(did)

        part1 = self.col.tr.importing_note_added(count=len(new))
        part2 = self.col.tr.importing_note_updated(count=self.updateCount)
        if self.importMode == UPDATE_MODE:
            unchanged = dupeCount - self.updateCount
        elif self.importMode == IGNORE_MODE:
            unchanged = dupeCount
        else:
            unchanged = 0
        part3 = self.col.tr.importing_note_unchanged(count=unchanged)
        self.log.append(f"{part1}, {part2}, {part3}.")
        self.log.extend(updateLog)
        self.total = len(self._ids)

    def newData(
        self, n: ForeignNote
    ) -> tuple[NoteId, str, NotetypeId, int, int, str, str, str, int, int, str]:
        id = self._nextID
        self._nextID = NoteId(self._nextID + 1)
        self._ids.append(id)
        self.processFields(n)
        # note id for card updates later
        for ord, c in list(n.cards.items()):
            self._cards.append((id, ord, c))
        return (
            id,
            guid64(),
            self.model["id"],
            int_time(),
            self.col.usn(),
            self.col.tags.join(n.tags),
            n.fieldsStr,
            "",
            0,
            0,
            "",
        )

    def addNew(
        self,
        rows: list[
            tuple[NoteId, str, NotetypeId, int, int, str, str, str, int, int, str]
        ],
    ) -> None:
        self.col.db.executemany(
            "insert or replace into notes values (?,?,?,?,?,?,?,?,?,?,?)", rows
        )

    def updateData(
        self, n: ForeignNote, id: NoteId, sflds: list[str]
    ) -> Updates | None:
        self._ids.append(id)
        self.processFields(n, sflds)
        if self._tagsMapped:
            tags = self.col.tags.join(n.tags)
            return (
                int_time(),
                self.col.usn(),
                n.fieldsStr,
                tags,
                id,
                n.fieldsStr,
                tags,
            )
        elif self.tagModified:
            tags = self.col.db.scalar("select tags from notes where id = ?", id)
            tagList = self.col.tags.split(tags) + self.tagModified.split()
            tags = self.col.tags.join(tagList)
            return (int_time(), self.col.usn(), n.fieldsStr, tags, id, n.fieldsStr)
        else:
            return (int_time(), self.col.usn(), n.fieldsStr, id, n.fieldsStr)

    def addUpdates(self, rows: list[Updates]) -> None:
        changes = self.col.db.scalar("select total_changes()")
        if self._tagsMapped:
            self.col.db.executemany(
                """
update notes set mod = ?, usn = ?, flds = ?, tags = ?
where id = ? and (flds != ? or tags != ?)""",
                rows,
            )
        elif self.tagModified:
            self.col.db.executemany(
                """
update notes set mod = ?, usn = ?, flds = ?, tags = ?
where id = ? and flds != ?""",
                rows,
            )
        else:
            self.col.db.executemany(
                """
update notes set mod = ?, usn = ?, flds = ?
where id = ? and flds != ?""",
                rows,
            )
        changes2 = self.col.db.scalar("select total_changes()")
        self.updateCount = changes2 - changes

    def processFields(self, note: ForeignNote, fields: list[str] | None = None) -> None:
        if not fields:
            fields = [""] * len(self.model["flds"])
        for c, f in enumerate(self.mapping):
            if not f:
                continue
            elif f == "_tags":
                note.tags.extend(self.col.tags.split(note.fields[c]))
            else:
                sidx = self._fmap[f][0]
                fields[sidx] = note.fields[c]
        note.fieldsStr = join_fields(fields)
        # temporary fix for the following issue until we can update the code:
        # https://forums.ankiweb.net/t/python-checksum-rust-checksum/8195/16
        if self.col.get_config_bool(Config.Bool.NORMALIZE_NOTE_TEXT):
            note.fieldsStr = unicodedata.normalize("NFC", note.fieldsStr)

    def updateCards(self) -> None:
        data = []
        for nid, ord, c in self._cards:
            data.append((c.ivl, c.due, c.factor, c.reps, c.lapses, nid, ord))
        # we assume any updated cards are reviews
        self.col.db.executemany(
            """
update cards set type = 2, queue = 2, ivl = ?, due = ?,
factor = ?, reps = ?, lapses = ? where nid = ? and ord = ?""",
            data,
        )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import enum
import logging
import mimetypes
import os
import re
import secrets
import sys
import threading
import traceback
from collections.abc import Callable
from dataclasses import dataclass
from errno import EPROTOTYPE
from http import HTTPStatus
from pathlib import Path

import flask
import stringcase
import waitress.wasyncore
from flask import Response, abort, request
from waitress.server import create_server

import aqt
import aqt.main
import aqt.operations
from anki import hooks
from anki.collection import OpChanges, OpChangesOnly, Progress, SearchNode
from anki.decks import UpdateDeckConfigs
from anki.scheduler.v3 import SchedulingStatesWithContext, SetSchedulingStatesRequest
from anki.utils import dev_mode
from aqt.changenotetype import ChangeNotetypeDialog
from aqt.deckoptions import DeckOptionsDialog
from aqt.operations import on_op_finished
from aqt.operations.deck import update_deck_configs as update_deck_configs_op
from aqt.progress import ProgressUpdate
from aqt.qt import *
from aqt.utils import aqt_data_path, show_warning, tr

# https://forums.ankiweb.net/t/anki-crash-when-using-a-specific-deck/22266
waitress.wasyncore._DISCONNECTED = waitress.wasyncore._DISCONNECTED.union({EPROTOTYPE})  # type: ignore

logger = logging.getLogger(__name__)
app = flask.Flask(__name__, root_path="/fake")


@dataclass
class LocalFileRequest:
    # base folder, eg media folder
    root: str
    # path to file relative to root folder
    path: str
    # collection media is untrusted user content; add-on web exports are not
    untrusted: bool = True


UNTRUSTED_MEDIA_CSP = "; ".join(
    (
        "default-src 'none'",
        "script-src 'none'",
        "connect-src 'none'",
        "object-src 'none'",
        "frame-src 'none'",
        "child-src 'none'",
        "base-uri 'none'",
        "form-action 'none'",
        "sandbox",
    )
)


def _editor_content_security_policy(port: int) -> str:
    # Only script-src is restricted here. Adding frame-src, object-src, img-src
    # etc. would break existing user content: YouTube embeds, dictionary iframes,
    # remote images, and SVG object tags.
    csp_paths = (
        f"http://127.0.0.1:{port}/_anki/",
        f"http://127.0.0.1:{port}/_addons/",
    )
    return "; ".join((f"script-src {' '.join(csp_paths)}",))


@dataclass
class BundledFileRequest:
    # path relative to aqt data folder
    path: str


@dataclass
class NotFound:
    message: str


DynamicRequest = Callable[[], Response]


class PageContext(enum.IntEnum):
    UNKNOWN = enum.auto()
    EDITOR = enum.auto()
    REVIEWER = enum.auto()
    PREVIEWER = enum.auto()
    CARD_LAYOUT = enum.auto()
    DECK_OPTIONS = enum.auto()
    # something in /_anki/pages/
    NON_LEGACY_PAGE = enum.auto()
    # Do not use this if you present user content (e.g. content from cards), as it's a
    # security issue.
    ADDON_PAGE = enum.auto()


@dataclass
class LegacyPage:
    html: str
    context: PageContext


class MediaServer(threading.Thread):
    _ready = threading.Event()
    daemon = True

    def __init__(self, mw: aqt.main.AnkiQt) -> None:
        super().__init__()
        self.is_shutdown = False
        # map of webview ids to pages
        self._legacy_pages: dict[int, LegacyPage] = {}

    def run(self) -> None:
        try:
            desired_host = os.getenv("ANKI_API_HOST", "127.0.0.1")
            desired_port = int(os.getenv("ANKI_API_PORT") or 0)
            self.server = create_server(
                app,
                host=desired_host,
                port=desired_port,
                clear_untrusted_proxy_headers=True,
            )
            logger.info(
                "Serving on http://%s:%s",
                self.server.effective_host,  # type: ignore[union-attr]
                self.server.effective_port,  # type: ignore[union-attr]
            )

            self._ready.set()
            self.server.run()

        except Exception:
            if not self.is_shutdown:
                raise

    def shutdown(self) -> None:
        self.is_shutdown = True
        sockets = list(self.server._map.values())  # type: ignore
        for socket in sockets:
            socket.handle_close()
        # https://github.com/Pylons/webtest/blob/4b8a3ebf984185ff4fefb31b4d0cf82682e1fcf7/webtest/http.py#L93-L104
        self.server.task_dispatcher.shutdown()

    def getPort(self) -> int:
        self._ready.wait()
        return int(self.server.effective_port)  # type: ignore

    def set_page_html(
        self, id: int, html: str, context: PageContext = PageContext.UNKNOWN
    ) -> None:
        self._legacy_pages[id] = LegacyPage(html, context)

    def get_page(self, id: int) -> LegacyPage | None:
        return self._legacy_pages.get(id)

    def get_page_html(self, id: int) -> str | None:
        if page := self.get_page(id):
            return page.html
        else:
            return None

    def get_page_context(self, id: int) -> PageContext | None:
        if page := self.get_page(id):
            return page.context
        else:
            return None

    def clear_page_html(self, id: int) -> None:
        try:
            del self._legacy_pages[id]
        except KeyError:
            pass


@app.route("/favicon.ico")
def favicon() -> Response:
    request = BundledFileRequest(os.path.join("imgs", "favicon.ico"))
    return _handle_builtin_file_request(request)


def _mime_for_path(path: str) -> str:
    "Mime type for provided path/filename."
    if path.endswith(".css"):
        # some users may have invalid mime type in the Windows registry
        return "text/css"
    elif path.endswith(".js") or path.endswith(".mjs"):
        return "application/javascript"
    else:
        # autodetect
        mime, _encoding = mimetypes.guess_type(path)
        return mime or "application/octet-stream"


def _text_response(code: HTTPStatus, text: str) -> Response:
    """Return an error message.

    Response is returned as text/plain, so no escaping of untrusted
    input is required."""
    resp = flask.make_response(text, code)
    resp.headers["Content-type"] = "text/plain"
    return resp


class UnsafePathException(Exception):
    def __init__(self, path: str):
        super().__init__(f"Invalid path: {path}")


def ensure_safe_path(base_dir: str | Path, path: str | Path) -> str:
    base_dir = os.path.realpath(base_dir)
    path = os.path.normpath(path)
    fullpath = os.path.abspath(os.path.join(base_dir, path))

    # protect against directory traversal: https://security.openstack.org/guidelines/dg_using-file-paths.html
    if not fullpath.startswith(base_dir + os.sep):
        raise UnsafePathException(path)
    return fullpath


_LOCALHOST_HOSTS = ("127.0.0.1", "localhost", "[::1]")

_ALLOWED_ORIGIN_PREFIXES = tuple(
    f"{scheme}{host}" for scheme in ("http://", "https://") for host in _LOCALHOST_HOSTS
)


def is_localhost_origin(origin: str) -> bool:
    for prefix in _ALLOWED_ORIGIN_PREFIXES:
        if (
            origin == prefix
            or origin.startswith(prefix + ":")
            or origin.startswith(prefix + "/")
        ):
            return True
    return False


def _handle_local_file_request(request: LocalFileRequest) -> Response:
    directory = request.root
    path = request.path
    try:
        isdir = os.path.isdir(os.path.join(directory, path))
    except ValueError:
        return _text_response(
            HTTPStatus.BAD_REQUEST, f"Path for '{directory} - {path}' is too long!"
        )

    fullpath = ensure_safe_path(directory, path)

    if isdir:
        return _text_response(
            HTTPStatus.FORBIDDEN,
            f"Path for '{directory} - {path}' is a directory (not supported)!",
        )

    try:
        mimetype = _mime_for_path(fullpath)
        if os.path.exists(fullpath):
            if fullpath.endswith(".css"):
                # caching css files prevents flicker in the webview, but we want
                # a short cache
                max_age = 10
            elif fullpath.endswith(".js"):
                # don't cache js files
                max_age = 0
            else:
                max_age = 60 * 60
            response = flask.send_file(
                fullpath,
                mimetype=mimetype,
                conditional=True,
                max_age=max_age,
                download_name="foo",  # type: ignore[call-arg]
            )
            if request.untrusted:
                # Prevent user-provided HTML/SVG from running as an active document.
                response.headers["Content-Security-Policy"] = UNTRUSTED_MEDIA_CSP
            return response
        else:
            print(f"Not found: {path}")
            return _text_response(HTTPStatus.NOT_FOUND, f"Invalid path: {path}")

    except Exception as error:
        if dev_mode:
            print(
                "Caught HTTP server exception,\n%s"
                % "".join(traceback.format_exception(*sys.exc_info())),
            )

        # swallow it - user likely surfed away from
        # review screen before an image had finished
        # downloading
        return _text_response(HTTPStatus.INTERNAL_SERVER_ERROR, str(error))


def _builtin_data(path: str) -> bytes:
    """Return data from file in aqt/data folder."""
    full_path = ensure_safe_path(aqt_data_path().parent, path)
    with open(full_path, "rb") as f:
        return f.read()


def _handle_builtin_file_request(request: BundledFileRequest) -> Response:
    path = request.path
    # do we need to serve the fallback page?
    immutable = "immutable" in path
    if path.startswith("sveltekit/") and not immutable:
        path = "sveltekit/index.html"
    mimetype = _mime_for_path(path)
    data_path = f"data/web/{path}"
    try:
        data = _builtin_data(data_path)
        response = Response(data, mimetype=mimetype)
        if immutable:
            response.headers["Cache-Control"] = "max-age=31536000"
        return response
    except FileNotFoundError:
        if dev_mode:
            print(f"404: {data_path}")
        resp = _text_response(HTTPStatus.NOT_FOUND, f"Invalid path: {path}")
        # we're including the path verbatim in our response, so we need to either use
        # plain text, or escape HTML characters to avoid reflecting untrusted input
        resp.headers["Content-type"] = "text/plain"
        return resp
    except Exception as error:
        if dev_mode:
            print(
                "Caught HTTP server exception,\n%s"
                % "".join(traceback.format_exception(*sys.exc_info())),
            )

        # swallow it - user likely surfed away from
        # review screen before an image had finished
        # downloading
        return _text_response(HTTPStatus.INTERNAL_SERVER_ERROR, str(error))


@app.route("/<path:pathin>", methods=["GET", "POST"])
def handle_request(pathin: str) -> Response:
    if os.environ.get("ANKI_API_HOST") != "0.0.0.0":
        host = request.headers.get("Host", "").lower()
        origin = request.headers.get("Origin", "").lower()
        allowed_hosts = tuple(f"{h}:" for h in _LOCALHOST_HOSTS)
        if not any(host.startswith(h) for h in allowed_hosts):
            logger.warning("denied non-local host: %s", host)
            abort(403)
        if origin and not is_localhost_origin(origin):
            logger.warning("denied non-local origin: %s", origin)
            abort(403)

    req = _extract_request(pathin)
    logger.debug("%s /%s", flask.request.method, pathin)

    try:
        if isinstance(req, NotFound):
            print(req.message)
            return _text_response(HTTPStatus.NOT_FOUND, f"Invalid path: {pathin}")
        elif callable(req):
            return _handle_dynamic_request(req)
        elif isinstance(req, BundledFileRequest):
            return _handle_builtin_file_request(req)
        elif isinstance(req, LocalFileRequest):
            return _handle_local_file_request(req)
        else:
            return _text_response(HTTPStatus.FORBIDDEN, f"unexpected request: {pathin}")
    except UnsafePathException as exc:
        return _text_response(HTTPStatus.FORBIDDEN, str(exc))


def is_sveltekit_page(path: str) -> bool:
    page_name = path.split("/")[0]
    return page_name in [
        "graphs",
        "congrats",
        "card-info",
        "change-notetype",
        "deck-options",
        "import-anki-package",
        "import-csv",
        "import-page",
        "image-occlusion",
    ]


def _extract_internal_request(
    path: str,
) -> BundledFileRequest | DynamicRequest | NotFound | None:
    "Catch /_anki references and rewrite them to web export folder."
    if is_sveltekit_page(path):
        path = f"_anki/sveltekit/_app/{path}"
    if path.startswith("_app/"):
        path = path.replace("_app", "_anki/sveltekit/_app")

    prefix = "_anki/"
    if not path.startswith(prefix):
        return None

    dirname = os.path.dirname(path)
    filename = os.path.basename(path)
    additional_prefix = None

    if dirname == "_anki":
        if flask.request.method == "POST":
            return _extract_collection_post_request(filename)
        elif get_handler := _extract_dynamic_get_request(filename):
            return get_handler

        # remap legacy top-level references
        base, ext = os.path.splitext(filename)
        if ext == ".css":
            additional_prefix = "css/"
        elif ext == ".js":
            if base in ("jquery-ui", "jquery", "plot"):
                additional_prefix = "js/vendor/"
            else:
                additional_prefix = "js/"
    # handle requests for vendored libraries
    elif dirname == "_anki/js/vendor":
        base, ext = os.path.splitext(filename)

        if base == "jquery":
            base = "jquery.min"
            additional_prefix = "js/vendor/"

        elif base == "jquery-ui":
            base = "jquery-ui.min"
            additional_prefix = "js/vendor/"

    if additional_prefix:
        oldpath = path
        path = f"{prefix}{additional_prefix}{base}{ext}"
        print(f"legacy {oldpath} remapped to {path}")

    return BundledFileRequest(path=path[len(prefix) :])


def _extract_addon_request(path: str) -> LocalFileRequest | NotFound | None:
    "Catch /_addons references and rewrite them to addons folder."
    prefix = "_addons/"
    if not path.startswith(prefix):
        return None

    addon_path = path[len(prefix) :]

    try:
        manager = aqt.mw.addonManager
    except AttributeError as error:
        if dev_mode:
            print(f"_redirectWebExports: {error}")
        return None

    try:
        addon, sub_path = addon_path.split("/", 1)
    except ValueError:
        return None
    if not addon:
        return None

    pattern = manager.getWebExports(addon)
    if not pattern:
        return None

    if re.fullmatch(pattern, sub_path):
        return LocalFileRequest(
            root=manager.addonsFolder(), path=addon_path, untrusted=False
        )

    return NotFound(message=f"couldn't locate item in add-on folder {path}")


def _extract_request(
    path: str,
) -> LocalFileRequest | BundledFileRequest | DynamicRequest | NotFound:
    if internal := _extract_internal_request(path):
        return internal
    elif addon := _extract_addon_request(path):
        return addon

    if not aqt.mw.col:
        return NotFound(message=f"collection not open, ignore request for {path}")

    path = hooks.media_file_filter(path)
    return LocalFileRequest(root=aqt.mw.col.media.dir(), path=path)


def congrats_info() -> bytes:
    if not aqt.mw.col.sched._is_finished():
        aqt.mw.taskman.run_on_main(lambda: aqt.mw.moveToState("overview"))
    return raw_backend_request("congrats_info")()


def get_deck_configs_for_update() -> bytes:
    return aqt.mw.col._backend.get_deck_configs_for_update_raw(request.data)


def update_deck_configs() -> bytes:
    # the regular change tracking machinery expects to be started on the main
    # thread and uses a callback on success, so we need to run this op on
    # main, and return immediately from the web request

    input = UpdateDeckConfigs()
    input.ParseFromString(request.data)

    def on_progress(progress: Progress, update: ProgressUpdate) -> None:
        if progress.HasField("compute_memory"):
            val = progress.compute_memory
            update.max = val.total_cards
            update.value = val.current_cards
            update.label = val.label
        elif progress.HasField("compute_params"):
            val2 = progress.compute_params
            # prevent an indeterminate progress bar from appearing at the start of each preset
            update.max = max(val2.total, 1)
            update.value = val2.current
            pct = str(int(val2.current / val2.total * 100) if val2.total > 0 else 0)
            label = tr.deck_config_optimizing_preset(
                current_count=val2.current_preset, total_count=val2.total_presets
            )
            if val2.reviews:
                reviews = tr.deck_config_percent_of_reviews(
                    pct=pct, reviews=val2.reviews
                )
            else:
                reviews = tr.qt_misc_processing()

            update.label = label + "\n" + reviews
        else:
            return
        if update.user_wants_abort:
            update.abort = True

    def on_success(changes: OpChanges) -> None:
        if isinstance(window := aqt.mw.app.activeModalWidget(), DeckOptionsDialog):
            window.reject()

    def handle_on_main() -> None:
        update_deck_configs_op(parent=aqt.mw, input=input).success(
            on_success
        ).with_backend_progress(on_progress).run_in_background()

    aqt.mw.taskman.run_on_main(handle_on_main)
    return b""


def get_scheduling_states_with_context() -> bytes:
    return SchedulingStatesWithContext(
        states=aqt.mw.reviewer.get_scheduling_states(),
        context=aqt.mw.reviewer.get_scheduling_context(),
    ).SerializeToString()


def set_scheduling_states() -> bytes:
    states = SetSchedulingStatesRequest()
    states.ParseFromString(request.data)
    aqt.mw.reviewer.set_scheduling_states(states)
    return b""


def import_done() -> bytes:
    def update_window_modality() -> None:
        if window := aqt.mw.app.activeModalWidget():
            from aqt.import_export.import_dialog import ImportDialog

            if isinstance(window, ImportDialog):
                window.hide()
                window.setWindowModality(Qt.WindowModality.NonModal)
                window.show()

    aqt.mw.taskman.run_on_main(update_window_modality)
    return b""


def import_request(endpoint: str) -> bytes:
    output = raw_backend_request(endpoint)()
    response = OpChangesOnly()
    response.ParseFromString(output)

    def handle_on_main() -> None:
        window = aqt.mw.app.activeModalWidget()
        on_op_finished(aqt.mw, response, window)

    aqt.mw.taskman.run_on_main(handle_on_main)

    return output


def import_csv() -> bytes:
    return import_request("import_csv")


def import_anki_package() -> bytes:
    return import_request("import_anki_package")


def import_json_file() -> bytes:
    return import_request("import_json_file")


def import_json_string() -> bytes:
    return import_request("import_json_string")


def search_in_browser() -> bytes:
    node = SearchNode()
    node.ParseFromString(request.data)

    def handle_on_main() -> None:
        aqt.dialogs.open("Browser", aqt.mw, search=(node,))

    aqt.mw.taskman.run_on_main(handle_on_main)

    return b""


def change_notetype() -> bytes:
    data = request.data

    def handle_on_main() -> None:
        window = aqt.mw.app.activeModalWidget()
        if isinstance(window, ChangeNotetypeDialog):
            window.save(data)

    aqt.mw.taskman.run_on_main(handle_on_main)
    return b""


def deck_options_require_close() -> bytes:
    def handle_on_main() -> None:
        window = aqt.mw.app.activeModalWidget()
        if isinstance(window, DeckOptionsDialog):
            window.require_close()

    # on certain linux systems, askUser's QMessageBox.question unsets the active window
    # so we wait for the next event loop before querying the next current active window
    aqt.mw.taskman.run_on_main(lambda: QTimer.singleShot(0, handle_on_main))
    return b""


def deck_options_ready() -> bytes:
    def handle_on_main() -> None:
        window = aqt.mw.app.activeModalWidget()
        if isinstance(window, DeckOptionsDialog):
            window.set_ready()

    aqt.mw.taskman.run_on_main(handle_on_main)
    return b""


def save_custom_colours() -> bytes:
    colors = [
        QColorDialog.customColor(i).name(QColor.NameFormat.HexArgb)
        for i in range(QColorDialog.customCount())
    ]
    aqt.mw.col.set_config("customColorPickerPalette", colors)
    return b""


post_handler_list = [
    congrats_info,
    get_deck_configs_for_update,
    update_deck_configs,
    get_scheduling_states_with_context,
    set_scheduling_states,
    change_notetype,
    import_done,
    import_csv,
    import_anki_package,
    import_json_file,
    import_json_string,
    search_in_browser,
    deck_options_require_close,
    deck_options_ready,
    save_custom_colours,
]


exposed_backend_list = [
    # CollectionService
    "latest_progress",
    # DeckService
    "get_deck_names",
    # I18nService
    "i18n_resources",
    # ImportExportService
    "get_csv_metadata",
    "get_import_anki_package_presets",
    # NotesService
    "get_field_names",
    "get_note",
    # NotetypesService
    "get_notetype_names",
    "get_change_notetype_info",
    # StatsService
    "card_stats",
    "get_review_logs",
    "graphs",
    "get_graph_preferences",
    "set_graph_preferences",
    # TagsService
    "complete_tag",
    # ImageOcclusionService
    "get_image_for_occlusion",
    "add_image_occlusion_note",
    "get_image_occlusion_note",
    "update_image_occlusion_note",
    "get_image_occlusion_fields",
    # SchedulerService
    "compute_fsrs_params",
    "compute_optimal_retention",
    "set_wants_abort",
    "evaluate_params_legacy",
    "get_optimal_retention_parameters",
    "simulate_fsrs_review",
    "simulate_fsrs_workload",
    # DeckConfigService
    "get_ignored_before_count",
    "get_retention_workload",
]


def raw_backend_request(endpoint: str) -> Callable[[], bytes]:
    # check for key at startup
    from anki._backend import RustBackend

    assert hasattr(RustBackend, f"{endpoint}_raw")

    return lambda: getattr(aqt.mw.col._backend, f"{endpoint}_raw")(request.data)


# all methods in here require a collection
post_handlers = {
    stringcase.camelcase(handler.__name__): handler for handler in post_handler_list
} | {
    stringcase.camelcase(handler): raw_backend_request(handler)
    for handler in exposed_backend_list
}


def _extract_collection_post_request(path: str) -> DynamicRequest | NotFound:
    if not aqt.mw.col:
        return NotFound(message=f"collection not open, ignore request for {path}")
    if handler := post_handlers.get(path):
        # convert bytes/None into response
        def wrapped() -> Response:
            try:
                if data := handler():
                    response = flask.make_response(data)
                    response.headers["Content-Type"] = "application/binary"
                else:
                    response = _text_response(HTTPStatus.NO_CONTENT, "")
            except Exception as exc:
                print(traceback.format_exc())
                response = _text_response(HTTPStatus.INTERNAL_SERVER_ERROR, str(exc))
            return response

        return wrapped
    else:
        return NotFound(message=f"{path} not found")


def _check_dynamic_request_permissions():
    if request.method == "GET":
        return

    def warn() -> None:
        show_warning(
            "Unexpected API access. Please report this message on the Anki forums."
        )

    # check content type header to ensure this isn't an opaque request from another origin
    if request.headers["Content-type"] != "application/binary":
        aqt.mw.taskman.run_on_main(warn)
        abort(403)

    # does page have access to entire API?
    if _have_api_access():
        return

    # whitelisted API endpoints for reviewer/previewer
    if request.path in (
        "/_anki/getSchedulingStatesWithContext",
        "/_anki/setSchedulingStates",
        "/_anki/i18nResources",
        "/_anki/congratsInfo",
    ):
        pass
    else:
        # other legacy pages may contain third-party JS, so we do not
        # allow them to access our API
        aqt.mw.taskman.run_on_main(warn)
        abort(403)


def _handle_dynamic_request(req: DynamicRequest) -> Response:
    _check_dynamic_request_permissions()
    try:
        return req()
    except Exception as e:
        return _text_response(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


def legacy_page_data() -> Response:
    id = int(request.args["id"])
    page = aqt.mw.mediaServer.get_page(id)
    if page:
        response = Response(page.html, mimetype="text/html")
        # Prevent JS in field content from being executed in the editor, as it would
        # have access to our internal API, and is a security risk.
        if page.context == PageContext.EDITOR:
            response.headers["Content-Security-Policy"] = (
                _editor_content_security_policy(aqt.mw.mediaServer.getPort())
            )
        return response
    else:
        return _text_response(HTTPStatus.NOT_FOUND, "page not found")


_APIKEY = secrets.token_urlsafe(32)


def _have_api_access() -> bool:
    return (
        request.headers.get("Authorization") == f"Bearer {_APIKEY}"
        or os.environ.get("ANKI_API_HOST") == "0.0.0.0"
    )


# this currently only handles a single method; in the future, idempotent
# requests like i18nResources should probably be moved here
def _extract_dynamic_get_request(path: str) -> DynamicRequest | None:
    if path == "legacyPageData":
        return legacy_page_data
    else:
        return None
